@@ -1,4 +1,6 @@
 from . import config
+from ._async import wrap_coroutine
+import copyreg
 import dateutil
 import datetime
 import functools
@@ -16,7 +18,6 @@ import inspect
 from dateutil import parser
 from dateutil.tz import tzlocal
 
-
 try:
     from maya import MayaDT
 except ImportError:
@@ -24,6 +25,7 @@ except ImportError:
 
 _TIME_NS_PRESENT = hasattr(time, 'time_ns')
 _MONOTONIC_NS_PRESENT = hasattr(time, 'monotonic_ns')
+_PERF_COUNTER_NS_PRESENT = hasattr(time, 'perf_counter_ns')
 _EPOCH = datetime.datetime(1970, 1, 1)
 _EPOCHTZ = datetime.datetime(1970, 1, 1, tzinfo=dateutil.tz.UTC)
 
@@ -31,10 +33,11 @@ real_time = time.time
 real_localtime = time.localtime
 real_gmtime = time.gmtime
 real_monotonic = time.monotonic
+real_perf_counter = time.perf_counter
 real_strftime = time.strftime
 real_date = datetime.date
 real_datetime = datetime.datetime
-real_date_objects = [real_time, real_localtime, real_gmtime, real_monotonic, real_strftime, real_date, real_datetime]
+real_date_objects = [real_time, real_localtime, real_gmtime, real_monotonic, real_perf_counter, real_strftime, real_date, real_datetime]
 
 if _TIME_NS_PRESENT:
     real_time_ns = time.time_ns
@@ -44,6 +47,10 @@ if _MONOTONIC_NS_PRESENT:
     real_monotonic_ns = time.monotonic_ns
     real_date_objects.append(real_monotonic_ns)
 
+if _PERF_COUNTER_NS_PRESENT:
+    real_perf_counter_ns = time.perf_counter_ns
+    real_date_objects.append(real_perf_counter_ns)
+
 _real_time_object_ids = {id(obj) for obj in real_date_objects}
 
 # time.clock is deprecated and was removed in Python 3.8
@@ -54,13 +61,6 @@ tz_offsets = []
 ignore_lists = []
 tick_flags = []
 
-# Python3 doesn't have basestring, but it does have str.
-try:
-    # noinspection PyUnresolvedReferences
-    _string_type = basestring
-except NameError:
-    _string_type = str
-
 try:
     # noinspection PyUnresolvedReferences
     real_uuid_generate_time = uuid._uuid_generate_time
@@ -81,23 +81,6 @@ try:
 except (AttributeError, ImportError):
     real_uuid_create = None
 
-try:
-    import copy_reg as copyreg
-except ImportError:
-    import copyreg
-
-try:
-    iscoroutinefunction = inspect.iscoroutinefunction
-    if sys.version_info < (3, 5):
-        from freezegun._async_coroutine import wrap_coroutine
-    else:
-        from freezegun._async import wrap_coroutine
-except AttributeError:
-    iscoroutinefunction = lambda x: False
-
-    def wrap_coroutine(*args):
-        raise NotImplementedError()
-
 
 # keep a cache of module attributes otherwise freezegun will need to analyze too many modules all the time
 _GLOBAL_MODULES_CACHE = {}
@@ -134,7 +117,7 @@ def _get_module_attributes_hash(module):
         module_dir = dir(module)
     except (ImportError, TypeError):
         module_dir = []
-    return '{}-{}'.format(id(module), hash(frozenset(module_dir)))
+    return f'{id(module)}-{hash(frozenset(module_dir))}'
 
 
 def _get_cached_module_attributes(module):
@@ -149,11 +132,6 @@ def _get_cached_module_attributes(module):
     return cached_attrs
 
 
-# Stolen from six
-def with_metaclass(meta, *bases):
-    """Create a base class with a metaclass."""
-    return meta("NewBase", bases, {})
-
 _is_cpython = (
     hasattr(platform, 'python_implementation') and
     platform.python_implementation().lower() == "cpython"
@@ -222,21 +200,51 @@ def fake_gmtime(t=None):
     return get_current_time().timetuple()
 
 
+def _get_fake_monotonic():
+    # For monotonic timers like .monotonic(), .perf_counter(), etc
+    current_time = get_current_time()
+    return (
+        calendar.timegm(current_time.timetuple()) +
+        current_time.microsecond / 1e6
+    )
+
+
+def _get_fake_monotonic_ns():
+    # For monotonic timers like .monotonic(), .perf_counter(), etc
+    current_time = get_current_time()
+    return (
+        calendar.timegm(current_time.timetuple()) * 1000000 +
+        current_time.microsecond
+    ) * 1000
+
+
 def fake_monotonic():
     if _should_use_real_time():
         return real_monotonic()
-    current_time = get_current_time()
-    return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
+
+    return _get_fake_monotonic()
+
+
+def fake_perf_counter():
+    if _should_use_real_time():
+        return real_perf_counter()
+
+    return _get_fake_monotonic()
+
 
 if _MONOTONIC_NS_PRESENT:
     def fake_monotonic_ns():
         if _should_use_real_time():
             return real_monotonic_ns()
-        current_time = get_current_time()
-        return (
-            calendar.timegm(current_time.timetuple()) * 1000000 +
-            current_time.microsecond
-        ) * 1000
+
+        return _get_fake_monotonic_ns()
+
+
+if _PERF_COUNTER_NS_PRESENT:
+    def fake_perf_counter_ns():
+        if _should_use_real_time():
+            return real_perf_counter_ns()
+        return _get_fake_monotonic_ns()
 
 
 def fake_strftime(format, time_to_format=None):
@@ -296,10 +304,7 @@ def date_to_fakedate(date):
                     date.day)
 
 
-class FakeDate(with_metaclass(FakeDateMeta, real_date)):
-    def __new__(cls, *args, **kwargs):
-        return real_date.__new__(cls, *args, **kwargs)
-
+class FakeDate(real_date, metaclass=FakeDateMeta):
     def __add__(self, other):
         result = real_date.__add__(self, other)
         if result is NotImplemented:
@@ -342,10 +347,7 @@ class FakeDatetimeMeta(FakeDateMeta):
         return issubclass(subclass, real_datetime)
 
 
-class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
-    def __new__(cls, *args, **kwargs):
-        return real_datetime.__new__(cls, *args, **kwargs)
-
+class FakeDatetime(real_datetime, FakeDate, metaclass=FakeDatetimeMeta):
     def __add__(self, other):
         result = real_datetime.__add__(self, other)
         if result is NotImplemented:
@@ -481,7 +483,7 @@ def _parse_tz_offset(tz_offset):
         return datetime.timedelta(hours=tz_offset)
 
 
-class TickingDateTimeFactory(object):
+class TickingDateTimeFactory:
 
     def __init__(self, time_to_freeze, start):
         self.time_to_freeze = time_to_freeze
@@ -491,7 +493,7 @@ class TickingDateTimeFactory(object):
         return self.time_to_freeze + (real_datetime.now() - self.start)
 
 
-class FrozenDateTimeFactory(object):
+class FrozenDateTimeFactory:
 
     def __init__(self, time_to_freeze):
         self.time_to_freeze = time_to_freeze
@@ -513,7 +515,7 @@ class FrozenDateTimeFactory(object):
         self.tick(delta=delta)
 
 
-class StepTickTimeFactory(object):
+class StepTickTimeFactory:
 
     def __init__(self, time_to_freeze, step_width):
         self.time_to_freeze = time_to_freeze
@@ -539,7 +541,7 @@ class StepTickTimeFactory(object):
         self.tick(delta=delta)
 
 
-class _freeze_time(object):
+class _freeze_time:
 
     def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds):
         self.time_to_freeze = _parse_time_to_freeze(time_to_freeze_str)
@@ -555,18 +557,19 @@ class _freeze_time(object):
     def __call__(self, func):
         if inspect.isclass(func):
             return self.decorate_class(func)
-        elif iscoroutinefunction(func):
+        elif inspect.iscoroutinefunction(func):
             return self.decorate_coroutine(func)
         return self.decorate_callable(func)
 
     def decorate_class(self, klass):
         if issubclass(klass, unittest.TestCase):
-            # If it's a TestCase, we assume you want to freeze the time for the
-            # tests, from setUpClass to tearDownClass
+            # If it's a TestCase, we freeze time around setup and teardown, as well
+            # as for every test case. This requires some care to avoid freezing
+            # the time pytest sees, as otherwise this would distort the reported
+            # timings.
 
-            # Use getattr as in Python 2.6 they are optional
-            orig_setUpClass = getattr(klass, 'setUpClass', None)
-            orig_tearDownClass = getattr(klass, 'tearDownClass', None)
+            orig_setUpClass = klass.setUpClass
+            orig_tearDownClass = klass.tearDownClass
 
             # noinspection PyDecorator
             @classmethod
@@ -574,10 +577,12 @@ class _freeze_time(object):
                 self.start()
                 if orig_setUpClass is not None:
                     orig_setUpClass()
+                self.stop()
 
             # noinspection PyDecorator
             @classmethod
             def tearDownClass(cls):
+                self.start()
                 if orig_tearDownClass is not None:
                     orig_tearDownClass()
                 self.stop()
@@ -585,13 +590,29 @@ class _freeze_time(object):
             klass.setUpClass = setUpClass
             klass.tearDownClass = tearDownClass
 
+            orig_setUp = klass.setUp
+            orig_tearDown = klass.tearDown
+
+            def setUp(*args, **kwargs):
+                self.start()
+                if orig_setUp is not None:
+                    orig_setUp(*args, **kwargs)
+
+            def tearDown(*args, **kwargs):
+                if orig_tearDown is not None:
+                    orig_tearDown(*args, **kwargs)
+                self.stop()
+
+            klass.setUp = setUp
+            klass.tearDown = tearDown
+
             return klass
 
         else:
 
             seen = set()
 
-            klasses = klass.mro() if hasattr(klass, 'mro') else [klass] + list(klass.__bases__)
+            klasses = klass.mro()
             for base_klass in klasses:
                 for (attr, attr_value) in base_klass.__dict__.items():
                     if attr.startswith('_') or attr in seen:
@@ -638,6 +659,7 @@ class _freeze_time(object):
 
         time.time = fake_time
         time.monotonic = fake_monotonic
+        time.perf_counter = fake_perf_counter
         time.localtime = fake_localtime
         time.gmtime = fake_gmtime
         time.strftime = fake_strftime
@@ -656,6 +678,7 @@ class _freeze_time(object):
             ('real_gmtime', real_gmtime, fake_gmtime),
             ('real_localtime', real_localtime, fake_localtime),
             ('real_monotonic', real_monotonic, fake_monotonic),
+            ('real_perf_counter', real_perf_counter, fake_perf_counter),
             ('real_strftime', real_strftime, fake_strftime),
             ('real_time', real_time, fake_time),
         ]
@@ -668,6 +691,10 @@ class _freeze_time(object):
             time.monotonic_ns = fake_monotonic_ns
             to_patch.append(('real_monotonic_ns', real_monotonic_ns, fake_monotonic_ns))
 
+        if _PERF_COUNTER_NS_PRESENT:
+            time.perf_counter_ns = fake_perf_counter_ns
+            to_patch.append(('real_perf_counter_ns', real_perf_counter_ns, fake_perf_counter_ns))
+
         if real_clock is not None:
             # time.clock is deprecated and was removed in Python 3.8
             time.clock = fake_clock
@@ -745,6 +772,7 @@ class _freeze_time(object):
 
             time.time = real_time
             time.monotonic = real_monotonic
+            time.perf_counter = real_perf_counter
             time.gmtime = real_gmtime
             time.localtime = real_localtime
             time.strftime = real_strftime
@@ -756,6 +784,9 @@ class _freeze_time(object):
             if _MONOTONIC_NS_PRESENT:
                 time.monotonic_ns = real_monotonic_ns
 
+            if _PERF_COUNTER_NS_PRESENT:
+                time.perf_counter_ns = real_perf_counter_ns
+
             if uuid_generate_time_attr:
                 setattr(uuid, uuid_generate_time_attr, real_uuid_generate_time)
             uuid._UuidCreate = real_uuid_create
@@ -779,16 +810,12 @@ class _freeze_time(object):
             return result
         functools.update_wrapper(wrapper, func)
 
-        # update_wrapper already sets __wrapped__ in Python 3.2+, this is only
-        # needed for Python 2.x support
-        wrapper.__wrapped__ = func
-
         return wrapper
 
 
 def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False, as_kwarg='',
                 auto_tick_seconds=0):
-    acceptable_times = (type(None), _string_type, datetime.date, datetime.timedelta,
+    acceptable_times = (type(None), str, datetime.date, datetime.timedelta,
              types.FunctionType, types.GeneratorType)
 
     if MayaDT is not None:
@@ -802,14 +829,14 @@ def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_ar
         raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')
 
     if isinstance(time_to_freeze, types.FunctionType):
-        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, auto_tick_seconds)
+        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds)
 
     if isinstance(time_to_freeze, types.GeneratorType):
-        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, auto_tick_seconds)
+        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds)
 
     if MayaDT is not None and isinstance(time_to_freeze, MayaDT):
         return freeze_time(time_to_freeze.datetime(), tz_offset, ignore,
-                           tick, as_arg)
+                           tick, as_arg, as_kwarg, auto_tick_seconds)
 
     if ignore is None:
         ignore = []
